(*
   SNU 4190.310
   by hgkang @ kaist
    + netj @ snu
*)

signature DPLUS = 
sig
  type id = string 
  type exp 
     = NUM of int 
     | TRUE 
     | FALSE 
     | UNIT                                   
     | VAR of id 
     | ADD of exp * exp 
     | SUB of exp * exp 
     | MUL of exp * exp 
     | DIV of exp * exp 
     | EQUAL of exp * exp 
     | LESS of exp * exp 
     | NOT of exp 
     | ASSIGN of id * exp 
     | SEQ of exp * exp 
     | IFTE of exp * exp * exp 
     | IFT of exp * exp 
     | WHILE of exp * exp 
     | LETV of id * exp * exp 
     | LETF of id * id * exp * exp
     | CALLV of id * exp 
     | READ of id 
     | WRITE of exp     
     | RECORD of (id * exp) list
     | FIELD of exp * id
     | ASSIGNF of exp * id * exp
  val print : exp -> unit     // pretty printer for D++ absyn 
  val run : exp -> unit
end  
  
structure Dplus : DPLUS =
struct
  type id = string 
  type exp 
     = NUM of int 
     | TRUE 
     | FALSE 
     | UNIT                                   
     | VAR of id 
     | ADD of exp * exp 
     | SUB of exp * exp 
     | MUL of exp * exp 
     | DIV of exp * exp 
     | EQUAL of exp * exp 
     | LESS of exp * exp 
     | NOT of exp 
     | ASSIGN of id * exp 
     | SEQ of exp * exp 
     | IFTE of exp * exp * exp 
     | IFT of exp * exp 
     | WHILE of exp * exp 
     | LETV of id * exp * exp 
     | LETF of id * id * exp * exp
     | CALLV of id * exp 
     | READ of id 
     | WRITE of exp     
     | RECORD of (id * exp) list
     | FIELD of exp * id
     | ASSIGNF of exp * id * exp
  
(*****************************************************)
(*   Pretty Printer for D++ Abstract Syntax Tree  *) 
(*****************************************************)
  val ps = print_string  val nl = print_newline
  fun indent(i) =   let
     fun it 0 = ()
       | it n = ps " "; it (n-1)  
  in nl(); it(i) end  
  fun pp (prog,i) = case prog of 
       NUM(n) => print_int n     
     | TRUE => ps "true"
     | FALSE => ps "false"
     | UNIT => ps "()"
     | VAR(x) => ps x
     | ADD(e1,e2) =>   ps "("; pp(e1,i); ps "+"; pp(e2,i); ps ")" 
     | SUB(e1,e2) =>   ps "("; pp(e1,i); ps "-"; pp(e2,i); ps ")"
     | MUL(e1,e2) =>   ps "("; pp(e1,i); ps "*"; pp(e2,i); ps ")"
     | DIV(e1,e2) =>   ps "("; pp(e1,i); ps "/"; pp(e2,i); ps ")"
     | EQUAL(e1,e2) => ps "("; pp(e1,i); ps "="; pp(e2,i); ps ")"
     | LESS(e1,e2) =>  ps "("; pp(e1,i); ps "<"; pp(e2,i); ps ")"
     | NOT(e) =>       ps "not("; pp(e,i); ps ")"
     | ASSIGN(id,e) => ps (id^" := "); pp(e,i) 
     | SEQ(e1,e2) => pp(e1,i); ps ";"; indent(i); pp(e2,i)
     | IFTE(e1,e2,e3)=> ps "if "; pp(e1,i); ps " then "; 
                       indent(i+1); pp(e2,i+1); 
                       indent(i); ps "else";
                       indent(i+1); pp(e3,i+1)
     | IFT(e1,e2) =>   ps "if "; pp(e1,i); ps " then "; 
                       indent(i+1); pp(e2,i+1) 
     | WHILE(e1,e2) => ps "while "; pp(e1,i); ps " do "; 
                       indent(i+1); pp(e2,i+1); 
                       indent(i); ps "end (* while *)"
     | LETV(id,e1,e2) => ps ("let var "^id^" := "); pp(e1,i); ps " in";
                        indent(i+1); pp(e2,i+1)
     | READ(id) => ps ("read "^id^"")
     | WRITE(e) => ps "write("; pp(e,i); ps ")"
     | LETF(id1,id2,e1,e2) => 
         ps("let proc "^id1^" ("^id2^") =");
         indent(i+1); pp(e1,i+1);
         indent(i); ps "in";
         indent(i+1); pp(e2,i+1);
         indent(i); ps "end"
     | CALLV(id,e) => ps ("call "^id ^ " ("); pp (e,i); ps ")"
     | RECORD [] => ps ("{}")
     | RECORD (h::t) =>
        let fun pf (id, e) = ps (id ^ " := "); pp (e, i)
        in  ps "{"; pf h; List.iter (fn p => ps ", "; pf p) t; ps "}"
        end
     | FIELD (e, id) => pp (e, i); ps ("." ^ id)
     | ASSIGNF (e1, id, e2) => pp (e1, i); ps ("." ^ id); pp (e2, i)
  fun print p = pp(p,0)

(*****************************************************)
(*               D++ Interpreter                  *) 
(*****************************************************)

  exception DError of string
  
  type value = B of bool | N of int | Unit | R of (string * address) list
   and address = Loc of int | Proc of id * exp * env
   and env = id -> address
  type memory = address -> value
  val empty_mem = fn x => raise DError " ޸ Դϴ"
  val empty_env = fn x => raise DError (x ^ "    Դϴ")
  fun store(mem,loc, v) = fn loc' => if loc'=loc then v else mem loc'     
  fun bind(env,id,addr) = fn x => if x=id then addr else env x
  val start_address = ref 0
  fun malloc() = Loc(start_address := !start_address+1; !start_address)
  fun get_fun (env,id) =
    (case env id of
        Proc(id,body,env) => (id,body,env)
      | _ => raise DError(id^" Լ ƴմϴ"))
  fun get_record R r = r
    | get_record _ = raise DError "ڵ尪 ʵ带   ֽϴ."
  fun get_field r x = List.assoc x r
        handle Not_found => raise DError (x ^ " 𸣴 ڵ ʵ ̸Դϴ")
                
 
  fun eval exp env m = case exp of
       NUM(n) => (N(n),m)
     | TRUE => (B(true),m)
     | FALSE => (B(false),m)
     | UNIT => (Unit,m)
     | VAR(x) => (m(env x), m)
     | ADD(e1,e2) => 
       let 
         val (v1,m1) = eval e1 env m
         val (v2,m2) = eval e2 env m1
       in        
         case (v1,v2) of 
           (N(n1),N(n2)) => (N(n1+n2), m2) 
         | _ => raise DError "+  纯 ̾մϴ" 
       end
     | SUB(e1,e2) => 
       let 
         val (v1,m1) = eval e1 env m
         val (v2,m2) = eval e2 env m1
       in        
         case (v1,v2) of 
           (N(n1),N(n2)) => (N(n1-n2), m2)
         | _ => raise DError "-  纯 ̾մϴ" 
       end
     | MUL(e1,e2) => 
       let 
         val (v1,m1) = eval e1 env m
         val (v2,m2) = eval e2 env m1
       in        
         case (v1,v2) of 
           (N(n1),N(n2)) => (N(n1*n2), m2)
         | _ => raise DError "*  纯 ̾մϴ" 
       end
     | DIV(e1,e2) => 
       let 
         val (v1,m1) = eval e1 env m
         val (v2,m2) = eval e2 env m1
       in        
         case (v1,v2) of 
           (N(n1),N(n2)) =>((N(n1/n2), m2) handle Division_by_zero 
                              => raise DError "0   ϴ")
         | _ => raise DError "/  纯 ̾մϴ" 
       end
     | EQUAL(e1,e2) => 
       let 
         val (v1,m1) = eval e1 env m
         val (v2,m2) = eval e2 env m1
         fun is_eq(B b1, B b2) = B(b1=b2)
           | is_eq(N n1, N n2) = B(n1=n2)
           | is_eq(Unit, Unit) = B(true)
           | is_eq(R r1, R r2) = B(r1=r2)
           | is_eq (_, _) = 
              raise DError "=  纯  Ÿ̾ մϴ"
       in        
         (is_eq(v1,v2), m2)
       end
     | LESS (e1, e2) => 
       let 
         val (v1,m1) = eval e1 env m
         val (v2,m2) = eval e2 env m1
       in        
         case (v1,v2) of 
           (N(n1),N(n2)) => (B(n1<n2), m2)
         | _ => raise DError "<  纯 ̾մϴ"
       end
     | NOT(e') => 
       let  
         val (v1,m1) = eval e' env m 
       in 
         case v1 of 
           (B b) => (B(not b),m1) 
         | _ => raise DError "not  operand bool Ÿ̾ մϴ"
       end 
     | ASSIGN(id,e') =>
       let 
         val (v,m1) = eval e' env m
       in
         (Unit, store(m1,env(id),v))
       end
     | SEQ(e1,e2) => 
       let
         val (_,m1) = eval e1 env m 
       in
         eval e2 env m1
       end       
     | IFTE(cond,e1,e2) => 
       let 
         val (v1,m1) = eval cond env m 
       in
         case v1 of 
           B(true) => eval e1 env m1 
         | B(false) => eval e2 env m1  
         | _ => raise DError "if condition bool  մϴ"
       end
     | IFT(cond,e1) => 
       let 
         val (v1,m1) = eval cond env m 
       in
         case v1 of 
           B(true) => eval e1 env m1 
         | B(false) => (Unit,m1)
         | _ => raise DError "if condition bool  մϴ"
       end
     | WHILE(c,e) => 
       let 
         val (v1,m1) = eval c env m  
       in
         case v1 of 
           B(true) => let val (_, m2) = eval e env m1
                      in eval exp env m2 end
         | B(false) => (Unit, m1)
         | _ => raise DError "while ǽ bool   մϴ"
       end
     | READ(id) =>
       let val m' = store (m, env(id), N(read_int())) 
                    handle Failure(int_of_string) => 
                           raise DError " Էµ  ֽϴ"
       in (m'(env id), m') end
     | WRITE(e) => 
       let 
         fun print_val(N(n)) = print_int(n); print_newline()
           | print_val _ = raise DError " µɼ ֽϴ."
         val (v,m1) = eval e env m    
       in
         (print_val(v);(Unit,m1))
       end          
     | LETV(id, init_exp, body) => 
       let 
         val (v,m') = eval init_exp env m
         val loc = malloc()
         val env' = bind(env,id,loc)
         val m'' = store(m',loc,v)
       in
         eval body env' m''
       end
     | LETF(f,x,e1,e2)=> eval e2 (bind(env,f,Proc(x,e1,env))) m
     | CALLV(f,e) =>
       let 
         val (v1, m1) = eval e env m
         val (x,body,env1) = get_fun(env,f)
         val env2 = bind(env1,f,Proc(x,body,env1))
         val loc = malloc()
         val proc_env = bind(env2,x,loc)    
       in 
         eval body proc_env (store(m1,loc,v1))
       end
     | RECORD (l) =>
        let fun evaleach (l, m) (x, e) =
                let val (v, m') = eval e env m in ((x, v)::l, m') end
            fun alloceach r m [] = (r, m)
              | alloceach r m ((x, v)::t) =
                let val loc = malloc ()
                    val r' = (x, loc) :: r
                    val m' = store (m, loc, v)
                in  alloceach r' m' t 
                end
            val (l', m') = List.fold_left evaleach ([], m) l
            val (r, m'') = alloceach [] m' l'
        in  (R r, m'')
        end
     | FIELD (e, x) =>
        let val (v, m') = eval e env m
            val r = get_record v
        in  (m' (get_field r x), m')
        end
     | ASSIGNF (e1, x, e2) =>
        let val (v1, m1) = eval e1 env m
            val r = get_record v1
            val (v2, m2) = eval e2 env m1
        in  (Unit, store (m2, (get_field r x), v2))
        end
       
  fun run prog = 
  	ignore(eval prog empty_env empty_mem) 
  	handle DError s => (print_string ("D++ Error : " ^ s); 
                           print_newline())
end
