(*
   SNU 4190.310
   by hgkang @ kaist
      netj @ snu
*)

fun run () =
 let
    val print_sm5_code = ref false  	
    and tailcall_sm5 = ref false
    and gc_sm5 = ref false
    val pp = print_string
    val dpp_file = ref ""
    val spec = 
         [("-pp", Arg.Set print_sm5_code, "Print Compiled SM5 code")
         ,("-tail", Arg.Set tailcall_sm5, "Using your own tail call SM5 module")
         ,("-gc", Arg.Set gc_sm5, "Using your own gc SM5 module")
         ]
    val usage =
         "Usage: run <options> <file> \nOptions are: "
    val _ =  (* argument parse *)
        Arg.parse
         spec
         (fn x => if Sys.file_exists x then dpp_file := x
                  else raise Arg.Bad("File does not exist"))
         usage
 in  if !dpp_file = "" then Arg.usage spec usage else
 let 
     val dpp_pgm = Parser.program
                       Lexer.start
                         (Lexing.from_channel (open_in !dpp_file))
 in if !tailcall_sm5 then 
      let val sm5_pgm = Tailcall_Compiler.trans dpp_pgm
      in  pp "== Running with your own tail call SM5 module ==\n";
          if !print_sm5_code then (
            pp "== Compiled SM 5 code ==\n"; 
            Tailcall_Sm5.print sm5_pgm;
            pp "\n"
          );
          pp "== D++ interpreter result ==\n";
          Dplus.run dpp_pgm;
          pp "== Compiled SM5 program execution result ==\n";
          Tailcall_Sm5.run sm5_pgm
      end
    else if !gc_sm5 then
      let val sm5_pgm = Gc_Compiler.trans dpp_pgm
      in  pp "== Running with your own gc SM5 module ==\n";
          if !print_sm5_code then (
            pp "== Compiled SM 5 code ==\n"; 
            Gc_Sm5.print sm5_pgm;
            pp "\n"
          );
          pp "== D++ interpreter result ==\n";
          Dplus.run dpp_pgm;
          pp "== Compiled SM5 program";
          pp "execution result ==\n";
          Gc_Sm5.run sm5_pgm
      end
    else
      let val sm5_pgm = Compiler.trans dpp_pgm
      in  if !print_sm5_code then (
            pp "== Compiled SM 5 code ==\n"; 
            Sm5.print sm5_pgm;
            pp "\n"
          );
          pp "== D++ interpreter result ==\n";
          Dplus.run dpp_pgm;
          pp "== Compiled SM5 program";
          pp "execution result ==\n";
          Sm5.run sm5_pgm
      end
  end
  end handle v => Error.handle_exn v 
  	
val _ = Printexc.catch run () 
